/**
 * 調用Worker中的方法。
 *
 * @param funcName 方法名稱。
 * @params params 方法參數。
 * @param workerName 是否在Worker中執行，默認為true。。
 * @returns
 */
export async function callWorkerFunction<K extends keyof WorkerFunctionMap>(funcName: K, params: Parameters<WorkerFunctionMap[K]>, workerName?: string)
{
    const _result: {
        /**
         * 返回結果
         */
        result?: ReturnType<WorkerFunctionMap[K]>,
        /**
         * 消耗時間
         */
        cost?: number,
    } = {};

    if (!workerFunctions.has(funcName))
    {
        console.warn(`未註冊WorkerFunction ${funcName} ，請使用方法 registerWorkerFunction 註冊WorkerFunction。`);

        return _result;
    }

    const start = Date.now();
    _result.result = await runWorkerFunction(workerName, funcName, params);
    _result.cost = Date.now() - start;

    return _result;
}

function getWorker(workerName: string)
{
    return workers.get(workerName);
}

function getFunction(funcName: string): Function
{
    return workerFunctions.get(funcName);
}

async function runWorkerFunction<K extends keyof WorkerFunctionMap>(workerName: string, funcName: K, params: Parameters<WorkerFunctionMap[K]>): Promise<ReturnType<WorkerFunctionMap[K]>>
{
    const worker = getWorker(workerName);
    if (worker)
    {
        const promise = new Promise<ReturnType<WorkerFunctionMap[K]>>((resolve) =>
        {
            worker.onmessage = ({ data: res }) =>
            {
                resolve(res);
            };
        });
        worker.postMessage({ funcName, params });

        return promise;
    }
    if (workerName)
    {
        console.warn(`未找到名称为 ${workerName} 的Worker，请使用registerWorker註冊Worker！强制运行在当前Worker中。`);
    }

    const func = getFunction(funcName);
    // eslint-disable-next-line prefer-spread
    const result = await func.apply(undefined, params);

    return result;
}

/**
 * 註冊可在Worker中執行的方法。
 *
 * @param funcName 方法名稱。
 * @param func 方法定義。
 *
 */
export function registerWorkerFunction<K extends keyof WorkerFunctionMap>(funcName: K, func: WorkerFunctionMap[K])
{
    if (workerFunctions.has(funcName))
    {
        console.warn(`重複註冊WorkerFunction ${funcName} `, func, workerFunctions.get(funcName));

        return;
    }
    workerFunctions.set(funcName, func);
}

/**
 * 註冊Worker。
 *
 * @param workerName Worker名稱。
 * @param worker Worker實例。
 *
 */
export function registerWorker(workerName: string, worker: Worker)
{
    if (workers.has(workerName))
    {
        console.warn(`重複註冊Worker ${workerName} `, worker, workers.get(workerName));

        return;
    }
    workers.set(workerName, worker);
}

/**
 * 注销Worker。
 *
 * @param workerName Worker名稱。
 * @param terminate 是否終止Worker。
 */
export function ungisterWorker(workerName: string, terminate = true)
{
    if (workers.has(workerName))
    {
        const worker: Worker = workers.get(workerName);
        workers.delete(workerName);
        if (terminate)
        {
            worker.terminate();
        }
    }
}

declare global
{
    /**
     * 申明所有需要在Worker中調用的方法。
     */
    interface WorkerFunctionMap
    {
    }
}

const workerFunctions = new Map();
const workers = new Map();
